{{
********************************************
* jvmVpCore - Virtual Peripheral functions *
********************************************
}}


{{

This is a Work In Progress (WIP).

This is not a request for coding critique. Suggestions for real improvements
rather than addressing subjective cosmetic cruft would be appreciated.

Hi, If you're reading this, you've downloaded a work that may or may not be
included in the Propeller JVM. I have had some experiences with assembly over
my 20 year software engineering career; they were mostly in "support of" large
embedded systems and multiprocessor designs to get C/C++ booted, for gdb, or
for ISRs handling.

My successful career leads me to believe I can cross any river or ocean, and
thus became involved with this after toying with propeller for some projects.

This is my first PASM attempt. It is not optimized, nor is it complete.
The TIMER, DAC, and PWM work; the ADC and UART devices are to be implemented.
I find it unlikely that all of this can happen in just one COG in 8.68us 8-)
Thus, this framework if possible may need replication to other COGS or be
done in a different way. It is expected that 2 instances of this object will
be required to implement full features of 6 virtual peripherals.

This code is not currently added to JVM build. I am encouraged to make it
public since it contains requirements documentation and a general framework
that may be included in the JVM. I have tested a variant of Timer with the
JVM. I will add this to JVM later pending some issue resolution.

All timings are verified on an O'scope.
Performance summary:

Device - Budget - Startup - R/W Update - Overhead - Device Time - Dev Count - Total - Headroom - Dev Budget
 Timer   8,680      320       1,400        220          400           6       5,440     3,240        940 
 Timer   8,680      320       1,400        220          400           3       3,580     5,100      2,100 
 DAC     8,680      320       1,400        220          620           6       6,760     1,920        940 
 DAC     8,680      320       1,400        220          620           3       4,240     4,440      2,100 
 PWM     8,680      320       1,400        220          980           6       8,920     (240)        940 
 PWM     8,680      320       1,400        220          980           3       5,320     3,360      2,100 

}}

{{
/**
 *
 * Implementing Virtual Peripherals (VP) requires a process to run in the background.
 * This is a good opportunity for Propeller, although it takes another COG out of
 * the user's hands. We don't have a strategy for the user to control a cog yet.
 * One possibility is extending the VP set to more cogs. Another is to provide a
 * way for a user binary to execute on a cog. These are TBD of course.
 *
 * The elements that need a background process are: ADC, DAC, PWM, Timer, and UART.
 * It is expected that one Propeller COG can behave as the VP thread which does
 * processing ever 8.68us. The goal is to have all VP's be processed in that time*.
 *
 * *Note: It is very likely that more than one COG should be used to realize the full
 * potential of the VP design. Having cog/hub shared memory access is very expensive
 * and may not allow time for even the best asm programmer to do everything in 8.68us.
 * Using a time-multiplex design would make all features possible, but this would reduce
 * the granularity of events.
 *
 * One can argue that certain functions such as the ADC do not need a timer process,
 * but the conversion can be made more accurate if done in time slices.  The DAC
 * must be done as part of a timer process since it is more or less just a PWM.
 *
 * Common VP design requirements:
 *
 * - Up to 6 VP should run at once.
 * - All VP elements should run every 8.68us.
 * - VP pin usage is mutually exclusive.
 *
 * - Requirement is to have up to 6 PWM or 6 ADC or 6 DAC or 6 Uart run at a time.
 *   The java code clearly uses only 1 timer at any given moment, the VP of the
 *   timer can be assigned and would replace one of the other VPs in the list above.
 *
 * - jvmNative method nm_core_CPU_uninstall stops a VP this is a good place to
 *   arrange notification for the cog to not update elements.
 *
 * Design requirements for ADC:
 *
 * - ADC will be initialized by native call.
 * - ADC value should be readable by user. Variable for this written by COG/read by java.
 * - ADC pins should be selectable one for input one for reference output.
 * - External RC network is required on the input pin for ADC to function.
 * - Input measurements are limited from 0 to 3.3VDC.
 * - Up to 6 ADC can be used at once; the code will update measurments on each 8.68us tick.
 *
 * Design requirements for DAC:
 *
 * - DAC will be initialized by native call.
 * - DAC should be writeable by user. Variable for this written by java, read by COG.
 * - DAC output pins should be selectable.
 * - External RC network is required on the output pin for DDC to function.
 * - Output measurements are limited from 0 to 3.3VDC.
 * - Up to 6 DAC can be used at once; the code will update output on each 8.68us tick.
 *
 * Design requirements for PWM:
 *
 * - PWM pins should be selectable.
 * - PWM will be initialized by native call.
 * - Each PWM pin's pulse cycle characteristics can be updated any time.
 * - Up to 6 PWM can be used at once; the code will update output on each 8.68us tick.
 *
 * Design requirements for Timer:
 *
 * - Timer should tick every 8.68us.
 * - Timer should start when initialzied by native call.
 * - Timer value should latch on demand from native call.
 * - The Propeller COG process main loop with wait is the basis for the timer.
 * - Timer variables will be incremented by COG and read by java when running.
 * - Timer latch message can be sent from java to COG and cleared by COG for java read.
 *
 * Design requirements for Uart:
 *
 * - Uart will operate in one direction per instance.
 * - Uart will buffer input or output at 256 bytes each.
 * - Uart receiver using HW flow control will stall sender when <= 16 RX buffer bytes are free.
 * - Uart receiver using HW flow control will remove back-pressure when > 16 bytes are free.
 * - Uart baud rate will be selectable from 600 to 57600 BPS.
 * - Uart stop bits will be selectable from 1 to 8.
 * - Uart level sense can be inverted for both data bits and control lines.
 * 
 * - Up to 6 uni-directional Uarts will be allowed.
 * 
 */

/**
 *
 * COG Initialization and Start/Stop design
 *
 * The jvmVpCore start method should be called at system startup (jd.init would be good).
 * Once started, the cog should run freely although stop is possible since all
 * initialization required for cog function should be put into start.
 *
 * VP Start/Stop. VPs are started by the given init or install method below.
 * Calls to jvmVpCore vp.startVP(addr, bank, type) can be added to these methods.
 *
 * nm_core_Uart_txInit(bank)
 * nm_core_Uart_rxInit(bank)
 * nm_core_Timer_init(bank)
 * nm_core_PWM_install(bank)
 * nm_core_DAC_install(bank)
 * nm_core_ADC_install(bank)
 *
 * All VPs are stopped individually by the nm_core_CPU_uninstall method.
 * Call to jvmVpCore vp.stopVP(bank) has been added.
 *
 */
 
/**
 * Timer design -- implemented. Timer can be in any VP slot.
 *
 * This is the easiest design as it involves "generally" incrementing the counter
 * on every pass of the waitcnt timeout in the main cog loop. More detail follows.
 *
 * Timer should start when initialzied by native call. At start time the nmTimer
 * registers will be cleared by java code. This init call is our opportunity to
 * know that the timer is reset, and reset our internal timer accordingly.
 *
 * The nmTimer register is written by the JVM/Java before startup. Afterwards
 * only the COG should write the register and JVM/Java read only the register
 * until the interface is disabled. When disabled, the COG will not access any register.
 * The nmLatchTimer register is written by cog, and read only by JVM/Java.
 * The nmLatchItMsg is set by JVM/Java to signal a timer latch; once written this
 * this value should not be written again until after it is read as clear by JVM/Java. 
 *
 * To prevent a deadlocks the COG and JVM/Java should obey the access rules.
 *
 * The cog will clear its internal timer on init and update latches on each tick.
 *
 * ' file stamp\core\Timer.java (offsets within a VP bank) 
 * nmPeriphISRVec = $00  '// offset for all VPs indicates the bank number VPTYPE*0x10
 *
 * This nmLatchTimer register set is not touched by cog according to jvmNative use.
 *
 * nmLatchTimer1  = $08  '// LSB of 32 bit timer latch
 * nmLatchTimer2  = $09  '
 * nmLatchTimer3  = $0A  '
 * nmLatchTimer4  = $0B  '// MSB of 32 bit timer latch ... not touched by cog
 *
 * This nmTimer register set is written once by Java/JVM at startup and accessed
 * as read or write by the process. Shitty design alert. At least java start
 * writes the counter before starting the timer. This is Little Endian like Prop.
 * 
 * nmTimer1       = $04  '// LSB of 32 bit timer ... read by cog / write by JVM
 * nmTimer2       = $05  '
 * nmTimer3       = $06  '
 * nmTimer4       = $07  '// MSB of 32 bit timer ... read by cog / write by JVM
 *
 * nmLatchItMsg   = $02
 * 
 * '///////// these registers are redefined to allow long access for Timer //////////
 */
 
/**
 * PWM design -- implemented. 6 PWM / COG.
 *
 * Only the initial state is known to not work at this time; it may be added later.
 *
 * The PWM is implemented according to the following rules:
 *
 * @param highTime the number of 8.68us the waveform should be high.
 * @param lowTime the number of 8.68us the waveform should be low.
 * @param highZero Boolean value (always true) that tells wether highTime != 0.
 * @param lowZero Boolean value (always true) that tells wether lowTime != 0.
 * The zero counter is niether set nor read by the COG.
 *
 * The registers below are read-only by COG and written by JVM/Java:
 *
 * 'file stamp\core\PWM.java (offsets within a VP bank) 
 * nmPWMPort     = $0E  '// 0 for PortA, 1 for PortB (reserved: 2 for PortC, 3 for PortD)
 * nmPWMPin      = $0F  '// bitmask (1 of 8 set)
 *
 * nmPWMhighTime = $0C
 * nmPWMlowTime  = $0D
 *
 * nmPWMhighZero, nmPWMlowZero, count, and state are ignored
 *
 * ///////// these registers are redefined to allow long access for PWM //////////
 *
 */
 
/**
 * ADC Design -- not implemented
 *
 * nmAdcValue    = $0b  '// current ADC value
 * nmInPort      = $0c  '// 0 for port0, 1 for port1
 * nmInPin       = $0d  '// bitmask (1 of 8 set)
 * nmOutPort     = $0e  '// 0 for port0, 1 for port1
 * nmOutPin      = $0f  '// bitmask (1 of 8 set)   
 */
 
/**
 * DAC Design -- implemented. Needs method optimization
 *
 * nmDacValue    = $0b  '// current ADC value
 * nmDacPort     = $0e  '// 0 for port0, 1 for port1
 * nmDacPin      = $0f  '// bitmask (1 of 8 set)   
 */
 
/**
 * UART Design -- TX implemented with restrictions.
 *                Backpressure not supported.
 *                Stop bit count selection fixed at large number for now ... can adjust later.
 *
 * UART Design -- RX not implemented.
 *
 * 'file stamp\core\Uart.java (offsets within a VP bank) 
 * nmUartDataH    = $02  '// transmit register highbyte
 * nmUartDataL    = $03  '// transmit register lowbyte 
 * nmUartInvert   = $05  '// 0 for normal, 1 for invert
 * nmUartBaud     = $06  '// 1000000/(bps * 4.34)
 * nmUartStopBits = $07  '// Tx only: 00000001 for 1 stop bit, 00000011 for 2 etc
 * nmUartPort     = $0C  '// 0 for PortA, 1 for PortB (reserved: 2 for PortC, 3 for PortD)
 * nmUartPin      = $0D  '// bit mask (1 of 8 set)
 * nmUartHSPort   = $0E  '// 0 for PortA, 1 for PortB (reserved: 2 for PortC, 3 for PortD)
 * nmUartHSPin    = $0F  '// bitmask (1 of 8 set)
 * nmUartBufferH  = $0A  '// address of 256 byte buffer (aligned to $XX00) 
 * nmUartBufferL  = $0B
 * nmUartHead     = $08  '//Index of the next free byte in the buffer. 
 * nmUartTail     = $09  '//Index of the oldest byte in the buffer.
 *                       '//If tailPointer == headPointer the buffer is empty.
 * 'Uart baud constants
 * speed600   = 192  'The speed value for 600 baud.
 * speed1200  = 96   'The speed value for 1200 baud.
 * speed2400  = 48   'The speed value for 2400 baud.
 * speed4800  = 24   'The speed value for 4800 baud.
 * speed7200  = 16   'The speed value for 7200 baud.
 * speed9600  = 12   'The speed value for 9600 baud.
 * speed14400 = 8    'The speed value for 14400 baud.
 * speed19200 = 6    'The speed value for 19200 baud.
 * speed28800 = 4    'The speed value for 28800 baud.
 * speed38400 = 3    'The speed value for 38400 baud.
 * speed57600 = 2    'The speed value for 57600 baud.
 */
  
}}


CON
  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000
  '_clkmode = xtal1 + pll8x
  '_xinfreq = 10_000_000

  VP_TICKCLK      = 692
  VP_TICKMUL      = 10_000_000
  VP_DEBUG_PIN    = 1             ' for debug only need to turn off for release
  
  VP_UTX_BITCNT   = $C0001
  ' 2^16+ is count, 2^0 is start bit, byte goes in bits 2^1-2^8
  ' Count $16 is emperically derived for BS2 to keep up at 38.4Kbps
  ' Count represents many stop bits.
  ' Should reduce from $16 to $A for 1 stop or $B for two stop.
  ' I have only used BS2 for now; could hook up Javelin and test later.
  VP_UART_VARCNT  = 7
   
CON
{ jump table variants - jump table has problems - delete
  WRITE         = $0100_0000 'write command
  READ          = $0200_0000 'read command
  ABSWRITE      = $0300_0000 'absolute addr write command
  ABSREAD       = $0400_0000 'absolute addr read command
  COMMAND       = $0500_0000 'command
}
{ bit switch variants }
  WRITE         = $0100_0000 'write command
  READ          = $0200_0000 'read command
  ABSWRITE      = $0400_0000 'absolute addr write command
  ABSREAD       = $0800_0000 'absolute addr read command
  COMMAND       = $1000_0000 'command

  RESPONSE      = $00f0_0000 'to OR with return value to vpCommand
  ANYCMD        = WRITE | READ | ABSREAD | ABSWRITE | COMMAND
  ANYCMD_SHFT   = 24
  REGADR        = $0000_01ff ' one of 512 registers


CON ' COG pin control addresses
' It is not possible to set OUTA from one COG to another.
' This explains why DIRA[pin]~~ in spin init code would not work.
' Using these addresses and readAbs/writeAbs methods we can set values.
  INA_ADDR      = $1F2
  OUTA_ADDR     = $1F4
  DIRA_ADDR     = $1F6
  CTRA_ADDR     = $1F8
  FRQA_ADDR     = $1FA
  PHSA_ADDR     = $1FC
  VCFG_ADDR     = $1FE
  VSCL_ADDR     = $1FF

CON

  vpAsmAdcAICdiff = def#nmAdcInMask-def#nmAdcCount
  vpAsmAdcAOIdiff = def#nmAdcOutMask-def#nmAdcInMask
  vpAsmAdcAAOdiff = def#nmAdcAccu-def#nmAdcOutMask
  vpAsmAdcAVAdiff = def#nmAdcValue-def#nmAdcAccu
   
  vpAsmDacAPVdiff = def#nmDacPin-def#nmDacValue
  vpAsmDacAVAdiff = def#nmDacValue-def#nmDacAccu
   
  vpAsmPwmAPHdiff = def#nmPWMPin-def#nmPWMhighTime
  vpAsmPwmAPLdiff = def#nmPWMPin-def#nmPWMlowTime
  vpAsmPwmAPCdiff = def#nmPWMPin-def#nmPWMCount
  vpAsmPwmAHCdiff = def#nmPWMhighTime-def#nmPWMCount
  vpAsmPwmALCdiff = def#nmPWMlowTime-def#nmPWMCount
  vpAsmPwmALHdiff = def#nmPWMlowTime-def#nmPWMhighTime
   
  vpAsmUartAPnBdd = def#nmUartPin-def#nmUartBaud
  vpAsmUartABhBdd = def#nmUartBufferH-def#nmUartBaud
  vpAsmUartABlBhd = def#nmUartBufferL-def#nmUartBufferH
  vpAsmUartABdIvd = def#nmUartBaud-def#nmUartInvert
  vpAsmUartAHdBld = def#nmUartHead-def#nmUartBufferL
  vpAsmUartATlHdd = def#nmUartTail-def#nmUartHead
  vpAsmUartAPnTld = def#nmUartPin-def#nmUartTail
  vpAsmUartAPnBld = def#nmUartPin-def#nmUartBufferL
  vpAsmUartAFcPnd = def#nmUartHSpin-def#nmUartPin
   

VAR
  long  cog

VAR
  long vpCommand 
  long vpData          

CON
  UBUFF_LEN     = $FF ' should always be a "mask value"
  
VAR ' uart buffers
  'byte txbuff[UBUFF_LEN+1]    ' transmit test buffer
  'byte rxbuff[UBUFF_LEN+1]    ' receive  test buffer

CON  
  SDEBUG        = 0
  RXTEST        = 1
      
OBJ
  def: "jvmDefines"
  sio: "jvmSerial"
  'vga: "VGA_Text"
  'dbg: "PASDebug" '<---- Added for Debugger


PUB main | addr, pw, invert, sbits, str

{
  if 1'SDEBUG
    'sio.start(31,30,0,28800)
    'sio.rx ' uncomment to pause before demo
    'vga.start(16)
   
  'dbg.start(31,30,@entry)

  pw := 1
  invert := 0
  sbits  := 1
  start(5000000)

  startURX(def#nmPeriph0Bank, pinmap(1), pinmap(2), def#speed9600, sbits, invert, @rxbuff) ' errors
  'startUTX(def#nmPeriph1Bank, pinmap(0), pinmap(2), def#speed19200, sbits, invert, @txbuff)
  uartRxTest(def#nmPeriph0Bank)
}
  
{
  startUTX(def#nmPeriph1Bank, pinmap(0), pinmap(2), def#speed9600, sbits, invert, @txbuff)
  'uartRxTest(def#nmPeriph0Bank)
  'uartEchoTest(def#nmPeriph0Bank, def#nmPeriph1Bank)
  'jvp.startUTX(def#nmPeriph0Bank, pinmap(3), pinmap(2), def#speed19200, sbits, invert, @txbuff)
  'uartTxTest(def#nmPeriph1Bank)
  repeat
    str := string("UUU*")
    'str := string("Hello World",$d,$a)
    vga.dec(strsize(str))
    vga.str(string(" tl "))
    vga.hex(readReg(def#nmPeriph1Bank + def#nmUartTail),4)
    uartTx(def#nmPeriph1Bank, str)
    vga.str(string(" hd "))
    vga.hex(readReg(def#nmPeriph1Bank + def#nmUartHead),4)
    vga.out($d)
    pause(100)

  'startADC(def#nmPeriph0Bank, pinmap(1), pinmap(2))
  'adcTest(def#nmPeriph0Bank)
  
  'startVP(def#nmPeriph0Bank, def#VP_TIMER)
  'startVP(def#nmPeriph1Bank, def#VP_TIMER)
  
  'startPWM(def#nmPeriph0Bank, pinmap(1), pw, pw)
  'startPWM(def#nmPeriph1Bank, pinmap(2), pw, pw)

  'startURX(def#nmPeriph0Bank, pinmap(1), pinmap(2), def#speed600, sbits, invert, @rxbuff)   ' untested
  'startURX(def#nmPeriph0Bank, pinmap(1), pinmap(2), def#speed1200, sbits, invert, @rxbuff)  ' works
  'startURX(def#nmPeriph0Bank, pinmap(1), pinmap(2), def#speed2400, sbits, invert, @rxbuff)  ' works
  'startURX(def#nmPeriph0Bank, pinmap(1), pinmap(2), def#speed4800, sbits, invert, @rxbuff)  ' untested
  'startURX(def#nmPeriph0Bank, pinmap(1), pinmap(2), def#speed9600, sbits, invert, @rxbuff)  ' works
  'startURX(def#nmPeriph0Bank, pinmap(1), pinmap(2), def#speed19200, sbits, invert, @rxbuff) ' works
  'startURX(def#nmPeriph0Bank, pinmap(1), pinmap(2), def#speed38400, sbits, invert, @rxbuff) ' errors
  'startURX(def#nmPeriph0Bank, pinmap(0), pinmap(2), def#speed19200, sbits, invert, @rxbuff) ' errors
  startUTX(def#nmPeriph1Bank, pinmap(1), pinmap(2), def#speed19200, sbits, invert, @txbuff)
  uartRxTest(def#nmPeriph0Bank)
  'uartEchoTest(def#nmPeriph0Bank, def#nmPeriph1Bank)
  'startUTX(def#nmPeriph0Bank, pinmap(3), pinmap(2), def#speed19200, sbits, invert, @txbuff)
  'uartTxTest(def#nmPeriph1Bank)
  
  'startUTX(def#nmPeriph0Bank, pinmap(1), pinmap(2), def#speed57600, sbits, invert, @txbuff)
  'startUTX(def#nmPeriph1Bank, pinmap(1), 0, def#speed57600, sbits, invert, @txbuff)

  'startDAC(def#nmPeriph0Bank, pinmap(1), 50)
  'startDAC(def#nmPeriph1Bank, pinmap(2), 200)

  'startPWM(def#nmPeriph0Bank, pinmap(1), pw, pw)
  'startPWM(def#nmPeriph1Bank, pinmap(2), pw, pw)
  'startPWM(def#nmPeriph0Bank, pinmap(1), pw, pw)
  'startPWM(def#nmPeriph1Bank, pinmap(2), pw+1, pw)


  'startVP(def#nmPeriph0Bank, def#VP_TIMER)
  addr := def#nmPeriph0Bank
  
  'startVP(def#nmPeriph1Bank, def#VP_TIMER)
  repeat
    writeReg(addr+def#nmLatchTimer1, readReg(addr+def#nmTimer1))

  repeat
    writeReg(addr+def#nmLatchTimer1, readReg(addr+def#nmTimer1))
    pause(100)


  repeat  
  addr := def#nmPeriph0Bank

  ' use repeat reg access during test to ensure we have room for user service    
  if SDEBUG
    repeat
      sio.out($d)
      writeReg(addr+def#nmLatchTimer1, readReg(addr+def#nmTimer1))
      sio.hex(readReg(addr+def#nmTimer1),8)
      sio.out(" ")
      sio.hex(readReg(addr+def#nmLatchTimer1),8)
      pause(200)
    repeat
      readPWM(def#nmPeriph0Bank)
      readPWM(def#nmPeriph1Bank)
      readPWM(def#nmPeriph2Bank)
      readPWM(def#nmPeriph3Bank)
      readPWM(def#nmPeriph4Bank)
      readPWM(def#nmPeriph5Bank)
      pause(10000)
  else
    repeat
      writeReg(addr+def#nmLatchTimer1, readReg(addr+def#nmTimer1))
      'readReg(addr+def#nmTimer1)
      'writeReg(addr+def#nmLatchTimer1, 0)
}


PUB start(inf)
  if cog <> 0
    return cog ' dont start more than once without stop first
  cog := cognew(@entry, @vpCommand) + 1
  return cog
    
PUB stop
  if cog
    cogstop(cog~ - 1)
  cog := 0

PUB pinmap(num)
  result := 1 << num
  
PUB writeReg(address,value) | mbox
{{ write value to vpRambank register in VP cog }}

  address -= def#nmPeriph0Bank
  address := RESPONSE | WRITE | address

  if SDEBUG
    sio.tx(def#kComError)
    sio.tx(def#kErrJvm)
    sio.str(string($a,$d,"WRR "))
    sio.hex(address,8)
    sio.out(" ")
    sio.hex(value,8)
    sio.out(" ")
    sio.dec(value)
    sio.tx(def#kFlag)

  vpData    := value
  vpCommand := address ' command

  repeat while vpCommand <> 0
       
{
  long[@vpCommand+cid+1]       := value
  long[@vpCommand+cid]         := address ' command

  repeat while long[@vpCommand+cid] <> 0             
    sio.out(" ")
    sio.hex(long[@vpCommand+cid],8)
    sio.out($d)
    pause(100)
}  
     
PUB readReg(address): rc
{{ read value from vpRambank register in VP cog }}

  address -= def#nmPeriph0Bank
  address := RESPONSE | READ  | address

  if SDEBUG
    sio.tx(def#kComError)
    sio.tx(def#kErrJvm)
    sio.str(string($a,$d,"RDR "))
    sio.hex(address,8)
    'sio.tx(def#kFlag)

  vpCommand := address
  repeat while vpCommand <> 0
  result := vpData
  
  if SDEBUG
    'sio.tx(def#kComError)
    'sio.tx(def#kErrMessage)
    sio.out(" ")
    sio.hex(result,8)
    sio.out(" ")
    sio.dec(result)
    sio.tx(def#kFlag)
{
  vpCommand[cid*2] := address
  repeat while long[@vpCommand+cid] <> 0

  result := long[@vpCommand+cid+1]

  if SDEBUG
    sio.hex(result,8)
}
     
PUB writeAbs(address,value) | saddr
{{ write value to vpRambank register in VP cog }}
  saddr := address
  address := RESPONSE | ABSWRITE | address
  vpData  := value
  vpCommand := address
  repeat while vpCommand <> 0
  
  if SDEBUG
    sio.tx(def#kComError)
    sio.tx(def#kErrJvm)
    sio.str(string($d,"WRA "))
    sio.hex(address,8)
    sio.out(" ")
    sio.hex(value,4)
    sio.tx(def#kFlag)
    'readAbs(saddr)

    
PUB readAbs(address): rc | saddr
{{ read value from vpRambank register in VP cog }}
  saddr := address
  address := RESPONSE | ABSREAD  | address
  vpCommand := address
  repeat while vpCommand <> 0
  result := vpData

  if SDEBUG
    sio.tx(def#kComError)
    sio.tx(def#kErrJvm)
    sio.str(string($d,"RDA "))
    sio.hex(address,8)
    sio.out(" ")
    sio.hex(result,4)
    sio.tx(def#kFlag)

    
PUB enableVp(address,value) | saddr
{{ tell cog to use address as jump vector. TBD value = 0 may disable later. }}
  saddr := address
  address -= def#nmPeriph0Bank
  address := RESPONSE | COMMAND  | address
  vpData  := value
  vpCommand := address
  repeat while vpCommand <> 0
  
  if SDEBUG
    sio.str(string($d,"ENA "))
    sio.hex(address,8)
    sio.out(" ")
    sio.hex(value,4)
  
  
PUB pause(ms)
  if ms < 10
    result := ((clkfreq / 1000000) * 9050) / 10
  elseif ms < 100    
    result := ((clkfreq / 1000000) * 9950) / 10
  elseif ms < 1000 
    result := ((clkfreq / 1000000) * 10090) / 10
  else
    result := (clkfreq / 1000)
  result *= ms
  waitcnt(cnt+result)
    

PUB startADC(bank, ipinbit, opinbit) | tmp
{{
/**
 * @fn startADC
 * Call from native init/install method to start
 * @param bank - bank number to start
 * @param ipinbit - bitmask 1 << num 
 * @param opinbit  - bitmask 1 << num
 */
 _ void startVP(int bank, int port, int pin, int hcnt, int lcnt);
}}
  ' we have to tell the VP COG to set these values
  tmp := readAbs(DIRA_ADDR)
  tmp &= !ipinbit
  writeAbs(DIRA_ADDR,tmp)
  tmp := readAbs(DIRA_ADDR)
  tmp |= opinbit
  writeAbs(DIRA_ADDR,tmp)

  writeReg(bank + def#nmAdcInMask, ipinbit)
  writeReg(bank + def#nmAdcOutMask, opinbit)

  startVP(bank, def#VP_ADC)

PUB startDAC(bank, pinbit, value) | tmp
{{
/**
 * @fn startDAC
 * Call from native init/install method to start
 * @param bank - bank number to start
 * @param port - 0..3
 * @param pin  - 0..7
 * @param value - number of 8.68us ticks to keep pulse high
 * @note - port will always be zero. pin will be saved as 2^n pin bit number. 
 */
 _ void startVP(int bank, int port, int pin, int hcnt, int lcnt);
}}
  ' we have to tell the VP COG to set these values
  tmp := readAbs(DIRA_ADDR)
  tmp |= pinbit
  writeAbs(DIRA_ADDR,tmp)
  tmp := readAbs(OUTA_ADDR)
  tmp |= pinbit         'start high. first transaction toggles low
  writeAbs(OUTA_ADDR,tmp)

  writeReg(bank + def#nmDacPin, pinbit)
  writeReg(bank + def#nmDacValue, value)

  startVP(bank, def#VP_DAC)


PUB startPWM(bank, pinbit, hcnt, lcnt) | tmp
{{
/**
 * @fn startPWM
 * Call from native init/install method to start
 * @param bank - bank number to start
 * @param port - 0..3
 * @param pin  - 0..7
 * @param hcnt - number of 8.68us ticks to keep pulse high
 * @param lcnt - number of 8.68us ticks to keep pulse low
 * @note - port will always be zero. pin will be saved as 2^n pin bit number. 
 */
 _ void startVP(int bank, int port, int pin, int hcnt, int lcnt);
}}
  ' we have to tell the VP COG to set these values
  tmp := readAbs(DIRA_ADDR)
  tmp |= pinbit
  writeAbs(DIRA_ADDR,tmp)
  tmp := readAbs(OUTA_ADDR)
  tmp |= pinbit         'start high. first transaction toggles low
  writeAbs(OUTA_ADDR,tmp)
  'used
  writeReg(bank + def#nmPWMPin, pinbit)
  writeReg(bank + def#nmPWMCount, 1) ' don't go negative
  writeReg(bank + def#nmPWMlowTime, lcnt)
  writeReg(bank + def#nmPWMhighTime, hcnt)
  'unused
  writeReg(bank + def#nmPWMState, 0)
  'writeReg(bank + def#nmPWMPort, port)

  startVP(bank, def#VP_PWM)

PUB startUTX(bank, pinbit, fcpinbit, speed, sbits, invrt, buff) | tmp
{{
/**
 * @fn startUTX
 * Call from native init/install method to start
 * @param bank - bank number to start
 * @param pinbit - bitmask 1 << num
 * @param fcpinbit - bitmask 1 << num
 * @param speed - numeric speed divisor
 * @param sbits - numeric stop bits
 * @param buff  - buffer address
 * @note - port will always be zero. pin will be saved as 2^n pin bit number. 
 */
 _ void startVP(int bank, int port, int pin, int hcnt, int lcnt);
}}
' we have to tell the VP COG to set these values
  tmp := readAbs(DIRA_ADDR)
  tmp |= pinbit
  writeAbs(DIRA_ADDR,tmp)
  'tmp := readAbs(OUTA_ADDR) ' let driver handle it
  'tmp |= pinbit
  'writeAbs(OUTA_ADDR,tmp)
  tmp := readAbs(DIRA_ADDR)
  tmp &= !fcpinbit
  writeAbs(DIRA_ADDR,tmp)

  writeReg(bank + def#nmUartBaud, speed<<1) ' we keep baud in movi bits
  writeReg(bank + def#nmUartStopBits, sbits) ' sbits are ignored for now
  writeReg(bank + def#nmUartBufferH, (buff >> 8) & $ff)
  writeReg(bank + def#nmUartBufferL, buff & $ff)
  writeReg(bank + def#nmUartHead, 0)
  writeReg(bank + def#nmUartTail, 0)
  writeReg(bank + def#nmUartInvert, invrt)
  writeReg(bank + def#nmUartHsPin, fcpinbit)
  writeReg(bank + def#nmUartHsPort, 0)
  writeReg(bank + def#nmUartPin, pinbit)
  writeReg(bank + def#nmUartPort, 0)

' writeReg(bank + def#nmUartDataBits, dbits)           ' 8 dbits always ?
' writeReg(bank + def#nmUartPrty, prty)                ' N parity always ?

' It is very unclear what nmUartDataH/L are for. java has a function called
' sendTransmitEmpty that checks if value is 0 and buffer queue count 0.
' The same information can be gathered with just buffer queue count.
' Will not be using nmUartDataH/L for transmitter.

  startVP(bank, def#VP_UART_TX)


PUB startURX(bank, pinbit, fcpinbit, speed, sbits, invrt, buff) |  tmp
{{
/**
 * @fn startURX
 * Call from native init/install method to start
 * @param bank - bank number to start
 * @param pinbit - pin << number
 * @param fcpinbit  - pin << number
 * @param speed - numeric speed divisor
 * @param sbits - numeric stop bits
 * @param buff  - buffer address
 * @note - port will always be zero. pin will be saved as 2^n pin bit number. 
 */
 _ void startVP(int bank, int port, int pin, int hcnt, int lcnt);
}}
  ' we have to tell the VP COG to set these values
  tmp := readAbs(DIRA_ADDR)
  tmp &= !pinbit
  writeAbs(DIRA_ADDR,tmp)

{ ' let higher level driver handle this ? 
}
  tmp := readAbs(DIRA_ADDR)
  tmp |= fcpinbit
  writeAbs(DIRA_ADDR,tmp)

  tmp := readAbs(OUTA_ADDR)
  tmp |= fcpinbit
  writeAbs(OUTA_ADDR,tmp)

  writeReg(bank + def#nmUartBaud, speed<<1) ' we keep baud in movi bits - need shift adjust for it.
  writeReg(bank + def#nmUartStopBits, sbits)
  writeReg(bank + def#nmUartBufferH, (buff >> 8) & $ff)
  writeReg(bank + def#nmUartBufferL, buff & $ff)
  writeReg(bank + def#nmUartHead, 0)
  writeReg(bank + def#nmUartTail, 0)
  writeReg(bank + def#nmUartInvert, invrt)
  writeReg(bank + def#nmUartHsPin, fcpinbit)
  writeReg(bank + def#nmUartHsPort, 0)
  writeReg(bank + def#nmUartPin, pinbit)
  writeReg(bank + def#nmUartPort, 0)
  
  startVP(bank, def#VP_UART_RX)


PUB startVP(bank, type)
{{
/**
 * @fn startVP
 * Call from native init/install method to start VP
 * @param bank - bank number to start
 * @param type - type of VPTypeEnum:
 *   VP_TIMER, VP_ADC, VP_DAC, VP_PWM, VP_UART_RX, VP_UART_TX
 * @note starting with type VP_NONE will not start any VP.
 */
 _ void startVP(int bank, VpTypeEnum type);
}}
  result := bank
  result >>= 4
  result &= $7 ' apparently bank 0 is 8. Num 7 ($70) is defined as "block", but not used
  if result => $0 and result =< $5
    enableVp(bank, type)

      
PUB stopVP(bank)
{{
/**
 * @fn stopVP
 * Call nm_core_CPU_uninstall to stop VP.
 * jvm assembler programmer should make an asm version of this.
 * @param bank - bank number to start
 */
 _ void stopVP(int bank);
}}
{{ call nm_core_CPU_uninstall to stop VP }}
  result := bank
  bank := bank - def#nmPeriph0Bank
  result >>= 4
  result &= $7 ' apparently bank 0 is 8. Num 7 ($70) is defined as "block", but not used
  if result => $0 and result =< $5
    enableVp(bank, 0)


CON
  tickout       = $10
  loopout       = $80
  debugout      = tickout | loopout
      
DAT
{{_____________________________________________________________________________
_______________________________________________________________________________

An assembly cog which performs Javelin-like Virtual Peripheral VP functions.
Executes every 8.68 usec if possible. Currently no over-limit protection.
VP_DEBUG_PIN references should be commented or replaced with NOP at some point.
_______________________________________________________________________________
_______________________________________________________________________________
}}
entry           org 0

{{_____________________________________________________________________________
Main entry for VP process
}}
'  --------- Debugger Kernel add this at Entry (Addr 0) ---------
'   long $34FC1202,$6CE81201,$83C120B,$8BC0E0A,$E87C0E03,$8BC0E0A
'   long $EC7C0E05,$A0BC1207,$5C7C0003,$5C7C0003,$7FFC,$7FF8
'  -------------------------------------------------------------- 
'______________________________________________________________________________
'
                mov     vpCmdPtr,   par                
                mov     vpDataPtr,  vpCmdPtr    'setup dataptr for read/write
                add     vpDataPtr,  #4          ' moved to clear loop pipe filler 

                mov     tBnk,       #time       ' clear uninitialized variables from time to tBnk
                movd    clearWrite, tBnk        ' set destination address for write 
clearLoop       sub     tBnk,       #1          ' decr address ... don't need to clear time
clearWrite      mov     0-0,        #0          ' clear address
                movd    clearWrite, tBnk        ' set new destination address
                cmp     tBnk,       #tBnk wc    ' are we there yet ?
                if_a    jmp         #clearLoop  ' if address > #tBnk, repeat loop              
                
                mov     time,   cnt             'record current time
                add     time,   period          'establish next period

                'or      dira,   #debugout       'set tickout pin to output remove for release
tickLoop        'xor     outa,   #tickout        'toggle tick output for frequency measurement remove for release
                call    #vpUpdate
                call    #vpUserCmd
                waitcnt time, period            'we made it in allocated time :P, wait for next tick
                jmp     #tickLoop               'loop to next tick

{{_____________________________________________________________________________
PRI function: vpUpdate(...)
Does all VP functions in much less than 8.68us ... if we're lucky :)
App. < 5.2us with user cmd updates for 6 timers.
This update relies on the user enableVp command to take the vp number
destined for rambank[bank+0] and convert it into a procedure address.
}}
vpUpdate        mov     vpAsmBank,  #vpRambank      ' reset bank
                movs    tickTable,  vpAsmBank       ' initial type read addr - last to first VP
                mov     vpAsmIndex, #def#MAX_COG_VPS' reset index              
tickTable       mov     vpAsmType,  0-0 wz
                'xor     outa,       #loopout
                nop
                if_nz   jmp         vpAsmType       ' don't add debug stuff here.
doVpJumpReturn  add     vpAsmBank,  #$10            ' increment bank pointer
                movs    tickTable,  vpAsmBank       ' new type read addr
                djnz    vpAsmIndex, #tickTable      
vpUpdate_ret    ret

{{_____________________________________________________________________________
PRI function: vpUserCmd(...)
Processes user commands from vpCmdPtr/vpDataPtr 
}}
'______________________________________________________________________________
'
'--------------------------------------------------------        
' get command
vpUserCmd       rdlong  vpAsmCmd,   vpCmdPtr    wz 'read command into vpAsmCmd
                'xor     outa,   #tickout
                if_z    jmp         #vpUserCmdExit
                rdlong  vpAsmData,  vpDataPtr   'read data into vpAsmData
                mov     vpAsmVpReg, vpAsmCmd
                and     vpAsmVpReg, #REGADR     'no shift reg already in 2^0
'--------------------------------------------------------        
' read from memory/write to user
vpUserRead      test    vpAsmCmd,   cmdREAD     wz
                if_z    jmp         #vpUserWrite
                add     vpAsmVpReg, #vpRambank
                movd    vpUserRdRg, vpAsmVpReg
                nop     'pipe filler
vpUserRdRg      wrlong  0-0,        vpDataPtr
                jmp     #vpUserDone
'--------------------------------------------------------
'read from user/write to vpRambank memory                                     
'upper bound spillover - should never happen
'writeReg limits address ... exits if address out of range.
vpUserWrite     test    vpAsmCmd,   cmdWRITE    wz
                if_z    jmp         #vpUserAbsRead
                add     vpAsmVpReg, #vpRambank
                movd    vpUserWrRg, vpAsmVpReg
                nop     'pipe filler
vpUserWrRg      mov     0-0, vpAsmData
                jmp     #vpUserDone
'--------------------------------------------------------        
' read from any cog memory/write to user
vpUserAbsRead   test    vpAsmCmd,   cmdABSREAD  wz
                if_z    jmp         #vpUserAbsWrite
                movd    vpUserAbsRdRg, vpAsmVpReg
                nop     'pipe filler
vpUserAbsRdRg   wrlong  0-0,        vpDataPtr
                jmp     #vpUserDone
'--------------------------------------------------------                                     
' read from user/write to any cog memory                                    
vpUserAbsWrite  test    vpAsmCmd,   cmdABSWRITE wz
                if_z    jmp         #vpUserCommand
                movd    vpUserAbsWrRg, vpAsmVpReg
                nop     'pipe filler
vpUserAbsWrRg   mov     0-0, vpAsmData
                jmp     #vpUserDone
'______________________________________________________________________________
'Command to set jump vector at vpRambank bank# base
'Put jump table method address in vprambank block bank offset 0
'if vpreg = rambank % $10. Reduces time for running update loop.
'Update loop goes from double indirection to single indirection.
vpUserCommand   test    vpAsmCmd,   cmdCOMMAND wz
                if_z    jmp         #vpUserDone
                mov     vpAsmSlot,  vpAsmVpReg
                shr     vpAsmSlot,  #4              'slot index
                mov     var1,       vpAsmVpReg
                shr     var1,       #1              'get index for VP slot table
                add     vpAsmVpReg, #vpRambank      'add rambank address
                movd    vpUserTable,vpAsmVpReg      'setup to read from user table index
                mov     vpAsmType,  vpAsmData
vpUserTable     mov     0-0,        vpAsmData wz    ' set table index in vpAsmVpReg
                mov     var2,       #vpCoreJumpTable'user jump table
                add     var2,       var1            'get table for this VP slot
                add     var2,       vpAsmType
                movd    vpUserJmpJmp,vpAsmVpReg     'modify jump instruction address location
                movs    vpUserJmpJmp,var2           'modify jump instruction address
                call    #vpUserInitProc                
vpUserJmpJmp    mov     0-0,        0-0
' last item in list - no jmp #vpUserDone here
'--------------------------------------------------------
vpUserDone      mov     vpAsmCmd,   #0
                wrlong  vpAsmCmd,   vpCmdPtr
vpUserCmdExit   'xor     outa,   #loopout
vpUserCmd_ret   ret

{{_____________________________________________________________________________
PRI function userInitProc
Initialize some variables for devices ?
input param: vpAsmType
}}
vpUserInitProc  mov     var1,   vpAsmType
                add     var1,   #vpUserInitTable
                movs    vpUserInitJump, var1
                add     vpAsmSlot,      #1 'init routine needs "one-start" slot num
vpUserInitJump  jmp     0-0
vpUserInitDone                
vpUserInitProc_ret  ret

{{_____________________________________________________________________________
  _____________________________________________________________________________
}}
{{_____________________________________________________________________________
PRI function: vpAsmNone(...)
For jump table entries not covered/valid in VP type/slot index 
}}
vpAsmNone       jmp     #doVpJumpReturn

{{_____________________________________________________________________________
PRI function: vpAsmNoInit(...)
For jump table initialization entries not covered/valid in VP type/slot index 
}}
vpAsmNoInit     jmp     #vpUserInitDone
     
{{_____________________________________________________________________________
PRI function: vpAsmTimer(...)
The function of the timer is to increment on every "system" tick.
Increment 32bits timer in bank specified by vpAsmBank
}}
'______________________________________________________________________________
'
'only called when vpType is TIMER
' using local storage and self-modifying add ....
' timer read/clear/latch commands must be used more time to do that now per command tick
' 2.8us

vpAsmTimer      mov     tBnk,   vpAsmBank
                add     tBnk,   #def#nmTimer4       ' little end first
                mov     var1,   #4
vpAsmTimerRdLp  shl     var3,   #8
                movs    vpAsmTimerRd,   tBnk    
                sub     tBnk,   #1
vpAsmTimerRd    mov     var2,   0-0
                or      var3,   var2
                djnz    var1,   #vpAsmTimerRdLp     
vpAsmTimerIncr  add     var3,   #1                  ' increment
                mov     var1,   #4
vpAsmTimerWrLp  mov     var2,   var3
                add     tBnk,   #1
                movd    vpAsmTimerWr,   tBnk    
                and     var2,   #$ff   
vpAsmTimerWr    mov     0-0,   var2
                shr     var3,   #8
                djnz    var1,   #vpAsmTimerWrLp
                jmp     #doVpJumpReturn

{ big endian version 
vpAsmTimer      mov     tBnk,   vpAsmBank
                add     tBnk,   #def#nmTimer1       ' big end first
                mov     var1,   #4
vpAsmTimerRdLp  shl     var3,   #8
                movs    vpAsmTimerRd,   tBnk    
                add     tBnk,   #1
vpAsmTimerRd    mov     var2,   0-0
                or      var3,   var2
                djnz    var1,   #vpAsmTimerRdLp     
vpAsmTimerIncr  add     var3,   #1                  ' increment
                mov     var1,   #4
vpAsmTimerWrLp  mov     var2,   var3
                sub     tBnk,   #1
                movd    vpAsmTimerWr,   tBnk    
                and     var2,   #$ff   
vpAsmTimerWr    mov     0-0,   var2
                shr     var3,   #8
                djnz    var1,   #vpAsmTimerWrLp
                jmp     #doVpJumpReturn
}
                      
'______________________________________________________________________________
'
{{_____________________________________________________________________________
PRI function: vpAsmAdc(...)
Deliver ADC value
Derived from Peter's routine. Value returned is inverted for some reason.
}}
vpAsmAdc        mov     tBnk, vpAsmBank
                add     tBnk, #def#nmAdcCount
                movs    vpAsmAdcCntRd,  tBnk
                movd    vpAsmAdcCntWr,  tBnk
                add     tBnk, #vpAsmAdcAICdiff
                movs    vpAsmAdcInMrd,  tBnk
                add     tBnk, #vpAsmAdcAOIdiff
                movs    vpAsmAdcOutMrd, tBnk
                add     tBnk, #vpAsmAdcAAOdiff
                movs    vpAsmAdcAccRd, tBnk
                movd    vpAsmAdcAccWr, tBnk
                movd    vpAsmAdcAccClr, tBnk
                add     tBnk, #vpAsmAdcAVAdiff
                movd    vpAsmAdcValueWr, tBnk
vpAsmAdcInMrd   mov     adcInMask,      0-0
vpAsmAdcOutMrd  mov     adcOutMask,     0-0
                test    adcInMask,      ina wz          ' if input <> 0
                if_nz   andn    outa,   adcOutMask      '   set output 0
                if_nz   jmp     #vpAsmAdcCntRd          '   increment sample count
                or      outa,           adcOutMask      ' else input == 0, set output 1
vpAsmAdcAccRd   mov     adcAccum,       0-0             ' get sample
                add     adcAccum,       #1
                cmp     adcAccum,       #256 wz         
                if_e    sub     adcAccum, #1            ' prevent overflow
vpAsmAdcAccWr   mov     0-0,            adcAccum
vpAsmAdcCntRd   mov     adcCount,       0-0             ' increment count
                add     adcCount,       #1              
vpAsmAdcCntWr   mov     0-0,            adcCount        ' save count
                and     adcCount,       #$FF wz         ' if end count, set flag & clear count 
                if_ne   jmp     #vpAsmAdcJmp
vpAsmAdcValueWr mov     0-0,            adcAccum        ' save accum to value
vpAsmAdcAccClr  mov     0-0,            #0              ' clear accumulator
vpAsmAdcJmp     jmp     #doVpJumpReturn

      
{{_____________________________________________________________________________
PRI function: vpAsmDac(...)
Produce DAC output
_______________________________________________________________________________
A DAC is nothing more than a PWM with a "duty cycle". The value is created by
the proportion of high/low time. Thus, an output voltage of 3.3 is obtained when
the duty cycle is near or at 100% high. And 0 is when the duty cycle is near or
at 100% low. With a 50% duty cycle the output voltage would be 1.65. This all
depends on the RC network of course. The R*C time constant should be near
255*8.68us or 2.2ms.
nmDacValue 
nmDacPort  
nmDacPin   
nmDacAccu  
}}
'______________________________________________________________________________
'Load values from relative addresses using  diffs.
'///////// This could use some optimization as in PWM below. /////////////
vpAsmDac        mov     tBnk,   vpAsmBank        'target bank
                add     tBnk,   #def#nmDacAccu  'add for count source
                movs    vpAsmDacAccRd,  tBnk    'set accu source
                movd    vpAsmDacAccWr,  tBnk    'set accu dest
                add     tBnk,   #vpAsmDacAVAdiff
vpAsmDacAccRd   mov     dacAccu,    0-0 wz      'get accum
                if_nz   jmp     #vpAsmDacSave           'if count non-zero keep state 'til next time
                movs    vpAsmDacValRd,  tBnk    'set value source
                add     tBnk,   #vpAsmDacAPVdiff
                movs    vpAsmDacPinRd,  tBnk    'set pin source  
vpAsmDacValRd   mov     dacValue,   0-0         'get value
vpAsmDacPinRd   mov     dacPinBit,  0-0         'get pin bits
'--------------------------------------------------------        
                xor     outa,       dacPinBit
                test    outa,       dacPinBit   wz ' MUST test this bit
                if_z    jmp     #vpAsmDacLoadLo
                mov     dacAccu,    dacValue    ' high count = value
                add     dacAccu,    #1          ' fudge. can't make timing w/more insts. to accomodate zero
                jmp     #vpAsmDacSave
vpAsmDacLoadLo  mov     dacAccu,    #256        ' low count  = 256 - value
                sub     dacAccu,    dacValue
'--------------------------------------------------------        
vpAsmDacSave    sub     dacAccu, #1             'decr accum
vpAsmDacAccWr   mov     0-0,    dacAccu         'set accum
'--------------------------------------------------------        
vpAsmDacJmp     jmp     #doVpJumpReturn
'______________________________________________________________________________
'
{{_____________________________________________________________________________
PRI function: vpAsmPwm(...)
Produce PWM output
PWM high/low is controlled by "state" ...
    byte[vpRambank + bank + def#nmPWMHtime + state] ' state = 0 for hi time, 1 for low time
}}
'______________________________________________________________________________
'
'load values from relative addresses using  diffs
'1.25us .... appears to be upper end for 6 VPs
'______________________________________________________________________________
'
vpAsmPwm        mov     tBnk,   vpAsmBank       'get bank addr
                add     tBnk,   #def#nmPWMCount 'add for count source
                movd    vpAsmPwmCount,  tBnk    'set count dest for sub
                movd    vpAsmPwmCtWr,   tBnk    'set count adr
vpAsmPwmCount   cmpsub  0-0,    #1  wz          'comp to 1 if dst > src, subtract 1
                if_nz   jmp     #vpAsmPwmJmp    'if count negative, reload with new PW
                add     tBnk,   #vpAsmPwmAPCdiff 'add to count for pin source
                movs    vpAsmPwmPin,    tBnk    'set pin source now
                movs    vpAsmPwmXor,    tBnk    'set pin source now
                sub     tBnk,   #vpAsmPwmAPLdiff 'go back to low time addr
vpAsmPwmPin     test    outa,   0-0  wz         'MUST TEST the bit ... xor wz wont do it
                if_z    sub     tBnk,   #vpAsmPwmALHdiff 'sub back LH to get high PWM addr
                movs    vpAsmPwmCtWr,   tBnk    'set time source to add back on count
vpAsmPwmXor     xor     outa,   0-0             'no need to cheat now. bigger budget.
vpAsmPwmCtWr    add     0-0,    0-0             'add PWM to count
vpAsmPwmJmp     jmp     #doVpJumpReturn      
'______________________________________________________________________________
'

{{_____________________________________________________________________________
PRI function: vpAsmUartInit(...)
Initialize variables for uart transmit & receive routines.

input param:    vpAsmUart0invt  First variable defined in a set of MAX_COG_VPS tx variables.
                                The variables define an array of structs used by transmitters.
input param:    vpAsmSlot       A one-start number defining the slot number
input param:    vpAsmVpReg      The address of base vpRambank + offset number bank

Other inputs are set of uart register offsets defined by user code.

Basic operation:
Since enable is last operation after setting variables, we use that event to
initialize uart properties. Then for each uart transmitter, "wrappers" are used
to adjust major variables for each channel.
}}
vpAsmUartInit   mov     var1,   #vpAsmUart0invt
                mov     var2,   vpAsmSlot           ' VP_UART_VARCNT vars for UartTx and UartRX
                sub     var1,   #VP_UART_VARCNT     ' adjust for djnz 
vpAsmUTinitLp   add     var1,   #VP_UART_VARCNT     ' add varcnt to variable slot times
                djnz    var2,   #vpAsmUTinitLp
                mov     tBnk,   vpAsmVpReg
                add     tBnk,   #def#nmUartInvert   'add invert
                movs    vpAsmUTIinvert, tBnk
                movd    vpAsmUTIinvert, var1
                add     tBnk,   #vpAsmUartABdIvd    'add baud
                add     var1,   #1
                movs    vpAsmUTIbaud,   tBnk
                movd    vpAsmUTIbaud,   var1
                add     tBnk,   #vpAsmUartABhBdd    'add buff
                add     var1,   #1
                movs    vpAsmUTIbuff,   tBnk
                movd    vpAsmUTIbuff,   var1
                movd    vpAsmUTIbuffS,  var1
                movd    vpAsmUTIbuffL,  var1
                add     tBnk,   #vpAsmUartABlBhd    'add buffL
                movs    vpAsmUTIbuffL,  tBnk
                add     tBnk,   #vpAsmUartAPnBld    'add pin
                add     var1,   #1
                movs    vpAsmUTIpin,    tBnk
                movd    vpAsmUTIpin,    var1
                add     tBnk,   #vpAsmUartAFcPnd    'add flow control pin
                add     var1,   #1
                movs    vpAsmUTIfcpin,  tBnk         
                movd    vpAsmUTIfcpin,  var1
vpAsmUTIinvert  mov     0-0,    0-0                
vpAsmUTIbaud    mov     0-0,    0-0                
vpAsmUTIbuff    mov     0-0,    0-0                 ' get buff upper
vpAsmUTIbuffS   shl     0-0,    #8                  ' shl buff upper 8x
vpAsmUTIbuffL   or      0-0,    0-0                 ' or in buff lower
vpAsmUTIpin     mov     0-0,    0-0
vpAsmUTIfcpin   mov     0-0,    0-0
                jmp     #vpUserInitDone


{{_____________________________________________________________________________
PRI function: vpAsmUartTx0(...)
This is a wrapper routine for UartTx.
}}
vpAsmUartTx0    mov     vpAsmUartbyte,  vpAsmUart0byte  ' load variables 
                mov     vpAsmUartinvt,  vpAsmUart0invt   
                mov     vpAsmUartbaud,  vpAsmUart0baud   
                mov     vpAsmUartbuff,  vpAsmUart0buff   
                mov     vpAsmUartpin,   vpAsmUart0pin    
                mov     vpAsmUartFcpin, vpAsmUart0fcp    
                call    #vpAsmUartTxCom                 ' jump to common routine
                mov     vpAsmUart0byte, vpAsmUartbyte
                jmp     #doVpJumpReturn

{{_____________________________________________________________________________
PRI function: vpAsmUartTx1(...)
This is a wrapper routine for UartTx.
}}
vpAsmUartTx1    mov     vpAsmUartbyte,  vpAsmUart1byte  ' load variables 
                mov     vpAsmUartinvt,  vpAsmUart1invt   
                mov     vpAsmUartbaud,  vpAsmUart1baud   
                mov     vpAsmUartbuff,  vpAsmUart1buff   
                mov     vpAsmUartpin,   vpAsmUart1pin    
                mov     vpAsmUartFcpin, vpAsmUart1fcp    
                call    #vpAsmUartTxCom                 ' jump to common routine
                mov     vpAsmUart1byte, vpAsmUartbyte
                jmp     #doVpJumpReturn

{{_____________________________________________________________________________
PRI function: vpAsmUartTxCom(...)
This procedure is the common core of the uart transmitters used to produce output.
Basic operation:
Since enable is last operation after setting variables, we use that event to
initialize uart properties in the vpAsmUartTxInit procedure. Then for each uart
transmitter, "wrappers" are used to adjust major variables for each channel.
Transmitter toggle the output bit on 8.68us*count where count = input baud periods.
Basically, a bit flies every baud const times.

First, the transmitter should check the bit inprogress count and buffer.
If bit in progress (set by the input baud counter variable),
    decrement our count and exit.
Else,
    setup the next bit for transmit by reading the "byte" variable.
If byte variable count (upper word of byte variable) is non-zero,
    set the state of new bit.
Else
    If head > tail and no backpressure,
        Load the byte from the tail,
        Advance the tail.
        Setup a bitstream in the "byte" variable consisting of a start bit @2^0,
            the byte lsb first from 2^1 to 2^8, the stop bits. The length of
            the bitstream is defined by the count in byte variable @2^16+ .
Do set bit
Do decrement count    
}}
vpAsmUartTxCom  mov     tBnk,           vpAsmBank       'get bank addr
                add     tBnk,           #def#nmUartHead
                movs    vpAsmUTheadRd,  tBnk
                movd    vpAsmUTheadWr,  tBnk            ' apparently, java wants "us" to AND the head
                add     tBnk,           #vpAsmUartATlHdd
                movs    vpAsmUTtailRd,  tBnk
                movd    vpAsmUTtailWr,  tBnk            ' common tail write instr
vpAsmUTheadRd   mov     vpAsmUhead,     0-0
vpAsmUTtailRd   mov     vpAsmUtail,     0-0

                test    vpAsmUartByte,  vpAsmBitCntCntM wz ' check for bit complete ... can this become cmpsub ?
                if_nz   jmp             #vpAsmUTcntDecr ' if cnt (bit in progress) exit
                mov     var1,           vpAsmUartbyte
                test    vpAsmUartbyte,  vpAsmBitCntAndM wz
                if_nz   jmp             #vpAsmUTsend    ' if bit number non-zero, do next bit
                
                cmp     vpAsmUtail,     vpAsmUhead wz
                if_z    jmp             #vpAsmUartTxJmp ' if nothing to tx, get out
                mov     var2,           vpAsmUartbuff   ' get buff
                add     var2,           vpAsmUtail      ' add tail
                rdbyte  vpAsmUartByte,  var2            ' read byte from hub mem ... no choice :<
                xor     vpAsmUartByte,  #$FF            ' invert data
                shl     vpAsmUartByte,  #1
                or      vpAsmUartByte,  vpAsmBitCntOrM  ' "OR" in start bit and bit count here to fill SMC nop slot
                add     vpAsmUtail,     #1              ' we are responsible for incrementing tail
                and     vpAsmUtail,     #UBUFF_LEN      ' mod LEN-1
                tjz     vpAsmUartInvt,  #vpAsmUTtailWr
                xor     vpAsmUartByte,  MASK_WORD       ' flip all byte/start/stop bits if invert
vpAsmUTtailWr   mov     0-0,            vpAsmUtail      ' save tail
                and     vpAsmUhead,     #UBUFF_LEN
vpAsmUTheadWr   mov     0-0,            vpAsmUhead      ' we are responsible for resetting head    

vpAsmUTsend     test    vpAsmUartByte,  #1 wz           ' var1 is left over from before jump
                muxnz   outa,           vpAsmUartpin    ' use mux optimization  ' 4 inst total, vs 6 for andn/xor
                mov     var1,           vpAsmUartByte   ' get bit count
                sub     var1,           vpAsmBitCntDecr wz' subtract bit count
                tjnz    var1,           #vpAsmUTmore

                tjz     vpAsmUartFcPin, #vpAsmUTnoFlow  ' skip section if FC pin var is 0
                tjnz    vpAsmUartInvt,  #vpAsmUTiFlow   ' if invert, fc on "zero"
                test    vpAsmUartFcPin, ina wz          ' get FC state to wz
                muxnz   var1,           #1
                jmp     #vpAsmUTtstFlow     
vpAsmUTiFlow    test    vpAsmUartFcPin, ina wz          ' get FC state to wz
                muxz    var1,           #1
vpAsmUTtstFlow  tjnz    var1,           #vpAsmUTnoFlow            
vpAsmUTFlow     mov     var2,           vpAsmUartBaud
                shl     var2,           #4              ' this delays next byte tx period
                movi    vpAsmUartByte,  var2
                jmp     #vpAsmUTcntDecr
vpAsmUTnoFlow

                mov     vpAsmUartbyte,  vpAsmBitCntCntD ' kill bit info. no delay
                jmp     #vpAsmUTcntDecr
vpAsmUTmore     and     vpAsmUartbyte,  MASK_WORD       ' clobber count
                shr     vpAsmUartbyte,  #1              ' drop out bit we just sent
                andn    var1,           MASK_WORD       ' make sure we don't or in old byteval
                or      vpAsmUartbyte,  var1            ' or in new bit count
                movi    vpAsmUartbyte,  vpAsmUartBaud   ' set new count ... baud is << 1
vpAsmUTcntDecr  sub     vpAsmUartByte,  vpAsmBitCntCntD
vpAsmUartTxJmp
vpAsmUartTxCom_ret      ret

{{_____________________________________________________________________________
PRI function: vpAsmUartRx0(...)
This is a wrapper routine for UartTx.
}}
vpAsmUartRx0    mov     vpAsmUartbyte,  vpAsmUart0byte  ' load variables 
                mov     vpAsmUartcount, vpAsmUart0count ' use for rxsync   
                mov     vpAsmUartinvt,  vpAsmUart0invt   
                mov     vpAsmUartbaud,  vpAsmUart0baud   
                mov     vpAsmUartbuff,  vpAsmUart0buff   
                mov     vpAsmUartpin,   vpAsmUart0pin    
                mov     vpAsmUartFcpin, vpAsmUart0fcp    
                call    #vpAsmUartRxCom                 ' jump to common routine
                mov     vpAsmUart0byte, vpAsmUartbyte
                mov     vpAsmUart0count,vpAsmUartcount   
                jmp     #doVpJumpReturn

{{_____________________________________________________________________________
PRI function: vpAsmUartRx1(...)
This is a wrapper routine for UartTx.
}}
vpAsmUartRx1    mov     vpAsmUartbyte,  vpAsmUart1byte  ' load variables 
                mov     vpAsmUartcount, vpAsmUart1count ' use for rxsync   
                mov     vpAsmUartinvt,  vpAsmUart1invt   
                mov     vpAsmUartbaud,  vpAsmUart1baud   
                mov     vpAsmUartbuff,  vpAsmUart1buff   
                mov     vpAsmUartpin,   vpAsmUart1pin    
                mov     vpAsmUartFcpin, vpAsmUart1fcp    
                call    #vpAsmUartRxCom                 ' jump to common routine
                mov     vpAsmUart1byte, vpAsmUartbyte
                mov     vpAsmUart1count,vpAsmUartcount   
                jmp     #doVpJumpReturn

{{_____________________________________________________________________________
PRI function: vpAsmUartRx(...)
Deliver Uart receive function

Receiver needs a start indication.
The stop bit count and start bit make up a sync bit-pattern "001".
When a sync pattern is detected set sync state in uart byte var InSync bit 2^23 (movi bit 0).
With a sync state bit <> 0, we are in receive state or "InSync".
Get bit.
If count
    decrement count and exit procedure.
If not InSync,
    then save bit state and advance sync ptr, check for "001" in sync var.
If not "001" exit procedure,
    else set InSync bit and fall through.
If InSync and bit cnt < 8, add bit to byte LSB to MSB (shl byte #1, or byte, bit)
    then exit procedure.
If InSync and bit cnt == 8, save byte to buff at head, advance head, clear InSync
    then exit procedure.
Flow control is handled in higher layer
}}                     
vpAsmUartRxCom  mov     tBnk,           vpAsmBank       'get bank addr
                add     tBnk,           #def#nmUartHead
                movs    vpAsmURheadRd,  tBnk
                movd    vpAsmURheadWr,  tBnk            ' common tail write instr
                add     tBnk,           #vpAsmUartATlHdd
                movs    vpAsmURtailRd,  tBnk
                movd    vpAsmURtailWr,  tBnk            ' tail update adjustment is done at higher level ?
vpAsmURheadRd   mov     vpAsmUhead,     0-0                          
vpAsmURtailRd   mov     vpAsmUtail,     0-0             ' don't need tail yet
                and     vpAsmUtail,     #$ff
vpAsmURtailWr   mov     0-0,            vpAsmUtail

vpAsmURisFlow   tjz     vpAsmUartFcpin, #vpAsmURnoFlow  ' skip section if FC var is 0
                mov     var1,           vpAsmUhead
                subabs  var1,           vpAsmUtail      ' get abs diff between head/tail
                cmp     var1,           #16 wc          ' if diff < 15, assert flow control based on invert state
                tjz     vpAsmUartInvt,  #vpAsmURiFlow
                if_nc   or      outa,   vpAsmUartFcpin
                if_c    andn    outa,   vpAsmUartFcpin
vpAsmURiFlow    if_nc   andn    outa,   vpAsmUartFcpin
                if_c    or      outa,   vpAsmUartFcpin
vpAsmURnoFlow

vpAsmUReceive   test    vpAsmUartPin,   ina wz          ' test value & convert to hex 1
                muxnz   rxbit,          #1              ' set bit 2^0 based on Z state
                cmp     vpAsmUartInvt,  #0 wz
                if_nz   xor rxbit,      #1
vpAsmURisRxSync test    vpAsmUartByte,  vpAsmBitRxSyncM wz  ' if rx or sync state, jump to count
                if_nz   jmp             #vpAsmURchkCnt     

vpAsmUgetSync   shl     vpAsmUartcount, #1              ' add to sync bitstream
                or      vpAsmUartcount, rxbit
                and     vpAsmUartcount, #$3
                cmp     vpAsmUartcount, #$1  wz         ' fixed at one stop bit for 3 VPs
                if_ne   jmp             #vpAsmUartRxCom_ret  ' no sync
vpAsmURinSync   or      vpAsmUartByte,  vpAsmBitInSyncM ' in sync set bit count
                jmp     #vpAsmURctReload

vpAsmURchkCnt   test    vpAsmUartByte,  vpAsmBitCntCntM wz ' check for bit complete ... can this become cmpsub ?
                if_nz   jmp             #vpAsmURcntDecr ' if cnt (bit in progress) exit
vpAsmURisRecv1  test    vpAsmUartByte,  vpAsmBitReceive wz ' check for receive state....
                if_nz   jmp             #vpAsmURcatbit    ' if cnt (bit in progress) exit

                cmp     vpAsmUhead,     vpAsmUtail  wc  ' if head < tail no room in buffer
                if_b    jmp             #vpAsmUartRxCom_ret 'no room in buffer ... get out

vpAsmURnewbyte  or      vpAsmUartByte,  vpAsmBitReceive ' in sync set receiving
                or      vpAsmUartByte,  vpAsmBitCntRrst ' in sync set bitcount
                movs    vpAsmUartByte,  #0              ' clear byte area
vpAsmURcatbit   mov     var2,           vpAsmUartByte   ' get bit count
                shr     var2,           #16             ' put in position
                mov     var1,           rxbit           ' save bit
                shl     var1,           #8              ' shift bit left for placement
                shr     var1,           var2            ' shift bit right to proper position
                or      vpAsmUartByte,  var1
                sub     vpAsmUartByte,  vpAsmBitCntDecr ' this works. could be optimized. keep for now.
vpAsmUtest8th   test    vpAsmUartByte,  vpAsmBitCntAndM wz  'continue old byte
                if_nz   jmp             #vpAsmURctReload

                cmp     vpAsmUartInvt,  #0 wz
                if_z    xor vpAsmUartByte, #$ff
                mov     var1,           vpAsmUartBuff
                add     var1,           vpAsmUHead
                wrbyte  vpAsmUartByte,  var1            ' write lowest byte to buffer
                add     vpAsmUhead,     #1
                and     vpAsmUhead,     #UBUFF_LEN
                andn    vpAsmUartByte,  vpAsmBitReceive ' receive done
                andn    vpAsmUartByte,  vpAsmBitInSyncM   
vpAsmURheadWr   mov     0-0,            vpAsmUhead      ' advance head pointer

vpAsmURctReload movi    vpAsmUartByte,  vpAsmUartBaud   ' reload count. baud is << 1 by startup gives us movi ability
vpAsmURcntDecr  sub     vpAsmUartByte,  vpAsmBitCntCntD
vpAsmUartRxCom_ret      ret      


{{_____________________________________________________________________________
  _____________________________________________________________________________
DAT ASM Variables
}}
MASK_WORD       long    $FFFF

cmdWRITE        long    WRITE       'write command
cmdREAD         long    READ        'read command
cmdABSWRITE     long    ABSWRITE    'write command
cmdABSREAD      long    ABSREAD     'read command
cmdCOMMAND      long    COMMAND     'command command - just used for enable/disable jump vector

vpAsmBitCntOrM  long    VP_UTX_BITCNT
vpAsmBitCntDecr long    $00010000       ' decrements bit count in upper 16 bits
vpAsmBitCntAndM long    $00FF0000       ' bit count mask
vpAsmBitCntRrst long    $00080000       ' receive reset value
vpAsmBitCntCntD long    $01000000       ' "cycle" count decrement
vpAsmBitCntCntM long    $FF000000       ' "cycle" count mask
vpAsmBitInSyncM long    $00008000
vpAsmBitReceive long    $00004000
vpAsmBitRxSyncM long    $0000C000

'jump table specifies a range of functions to call. A 0 (zero) entry means do not jump.
'
vpCoreJumpTable long    (@vpAsmNone     - @entry)/4 '0  zero is do nothing
                long    (@vpAsmTimer    - @entry)/4 '1
                long    (@vpAsmAdc      - @entry)/4 '2
                long    (@vpAsmDac      - @entry)/4 '3
                long    (@vpAsmPwm      - @entry)/4 '4
                long    (@vpAsmUartRx0  - @entry)/4 '5
                long    (@vpAsmUartTx0  - @entry)/4 '6
                long    (@vpAsmNone     - @entry)/4 '7

vpCoreJumpTab1  long    (@vpAsmNone     - @entry)/4 '0  zero is do nothing
                long    (@vpAsmTimer    - @entry)/4 '1
                long    (@vpAsmAdc      - @entry)/4 '2
                long    (@vpAsmDac      - @entry)/4 '3
                long    (@vpAsmPwm      - @entry)/4 '4
                long    (@vpAsmUartRx1  - @entry)/4 '5
                long    (@vpAsmUartTx1  - @entry)/4 '6
                long    (@vpAsmNone     - @entry)/4 '7

vpUserInitTable long    (@vpAsmNoInit   - @entry)/4 '0  zero is do nothing
                long    (@vpAsmNoInit   - @entry)/4 '1
                long    (@vpAsmNoInit   - @entry)/4 '2
                long    (@vpAsmNoInit   - @entry)/4 '3
                long    (@vpAsmNoInit   - @entry)/4 '4
                long    (@vpAsmUartInit - @entry)/4 '5
                long    (@vpAsmUartInit - @entry)/4 '6
                long    (@vpAsmNoInit   - @entry)/4 '7

'tickout         long    |< 25     'APIN=0
'infreq          long    5_000_000 'overloaded by user with start(freq)
period          long    VP_TICKCLK/2 * (VP_TICKMUL / 5_000_000) '115200 Hz (8.68S period)

vpCmdPtr        res     1 ' don't clear these par interface variables
vpDataPtr       res     1

tBnk            res     1 ' any location from here + 1 to time address is cleared at startup.
var1            res     1
var2            res     1
var3            res     1

vpAsmCmd        res     1
vpAsmData       res     1

vpAsmIndex      res     1         'the VP index getting an update
vpAsmBank       res     1
vpAsmType       res     1
vpAsmSlot       res     1
vpAsmVpReg      res     1

adcInMask       res     1
adcOutMask      res     1
adcAccum        res     1
adcCount        res     1

dacAccu         res     1
dacValue        res     1
dacPinBit       res     1

pwmPinBit       res     1
pwmState        res     1
pwmCount        res     1
pwmStAdr        res     1
pwmCtAdr        res     1
pwmHiAdr        res     1
pwmLoAdr        res     1

vpAsmUhead      res     1
vpAsmUtail      res     1
vpAsmUartbyte   res     1
vpAsmUartcount  res     1
vpAsmUartinvt   res     1
vpAsmUartbaud   res     1
vpAsmUartbuff   res     1
vpAsmUartpin    res     1
vpAsmUartFCpin  res     1

rxbit           res     1

'Up to XXX (3 for now) VPs can be used in this design because cpu design and speed
'Could be more generic if indirect addressing available processor ran at 300+ MHz :) 
'and using all variables dynamically such as the count/head/tail pointers, etc...
'to change the state of these variables requires restart on the vp

' order of these variables is important
vpAsmUart0byte  res     1
vpAsmUart0count res     1
vpAsmUart0invt  res     1
vpAsmUart0baud  res     1
vpAsmUart0buff  res     1
vpAsmUart0pin   res     1
vpAsmUart0fcp   res     1

' order of these variables is important
vpAsmUart1byte  res     1
vpAsmUart1count res     1
vpAsmUart1invt  res     1
vpAsmUart1baud  res     1
vpAsmUart1buff  res     1
vpAsmUart1pin   res     1
vpAsmUart1fcp   res     1

vpRambank       res     32 '[def#MAX_COG_VPS*16] 'registers for Virtual Peripherals (16 entries per VP)
time            res     1

{{_____________________________________________________________________________
  _____________________________________________________________________________
}}
                fit     $1f0